package server.persist;

import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.StreamCorruptedException;
import java.util.Vector;

public class PersistDataMessageQueue {

	private static int capacity = 100;
	private static String queueFileNamePrefix = ".que";

	private Vector queue;
	
	private String queueName;
	private String queueFileName;
	private File queueFile;
	
	private FileInputStream fileInStream;
	private FileOutputStream fileOutStream;
	private ObjectInputStream queueInStream;
	private ObjectOutputStream queueOutStream;
	
	protected PersistDataMessageQueue(String n) {
		queueName = n;
		queue = new Vector(capacity);
		initPersistStorage();
	}

	/**
	 * In case the queue has been created before, restore the queue state
	 * from the persist storage, i.e. a file.
	 * In case the queue has never been created, build persist storage for the 
	 * queue
	 */
	private void initPersistStorage() {
		queueFileName = new String(PersistDataMessageQueueManager.queueDirectoryName + "/" + queueName + queueFileNamePrefix);
		queueFile = new File(queueFileName);
		
		// in case queueFile exists, reading the file and restore the state
		if (queueFile.exists()) {
			try {
				fileInStream = new FileInputStream(queueFile);
				queueInStream = new ObjectInputStream(fileInStream);
				
				// read message data back to the queue
				queue = (Vector) queueInStream.readObject();
				
				queueInStream.close();
				fileInStream.close();
				
			} catch (IOException e) {
				System.err.println("Error: " + e.toString());
				if (e instanceof StreamCorruptedException) {
					System.err.println("Error: File content corrupted! Rebuild the file");
				}
				else if (e instanceof EOFException) {
					System.err.println("Error: File is empty! Rebuild the file");
					queue = new Vector();	
				}
				// delete the old file which is corrupted or empty
				try {
					if (queueInStream != null) queueInStream.close();
					if (fileInStream != null) fileInStream.close();
					queueFile.delete();
				} catch (IOException e2) {
					// TODO Auto-generated catch block
					e2.printStackTrace();
				}
			} catch (ClassNotFoundException e1) {
				System.err.println("Error: Fail to read Queue Object from the queue file");
			}
		}
		
		// initialization for the first time
		if (!queueFile.exists()) {
			try {
				// build a new file storage
				queueFile.createNewFile();
				queue = new Vector();
			} catch (IOException e) {
				System.err.println("Error: fail to create file " + queueFileName);
			}
		}
	}
	
	/**
	 * Destroy persist file storage when deleting the queue
	 */
	protected void destroyPersistStorage() {
		try {
			fileInStream.close();
			fileOutStream.close();

			queueInStream.close();
			queueOutStream.close();
			
			boolean result = queueFile.delete();
			if (!result) {
				System.err.println("Fail to delete the file");
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public String getName() {
		return queueName;
	}

	/**
	 * Get messages out of the FIFO Queue
	 * 
	 * @return null : if the queue is empty message : otherwise
	 */
	public String dequeue() {
		if (queue.isEmpty())
			return null;
		else {
			PersistDataMessage msg = (PersistDataMessage) queue.remove(0);
			String cont = msg.getContent();
			
			updatePersistStorage();
			
			return cont;
		}
	}

	/**
	 * Add messages into the FIFO Queue
	 * 
	 * @param String msg
	 * @return false : if the queue is full true : successfully add the message
	 */
	public boolean enqueue(String msg) {
		int size = queue.size();
		if (size >= capacity) {
			return false;
		} else {
			PersistDataMessage pmsg = new PersistDataMessage(msg, queueName);
			queue.addElement(pmsg);
			updatePersistStorage();
			
			return true;
		}
	}
	
	/**
	 * When there is new message incoming or old message outgoing, update the storage
	 */
	private void updatePersistStorage() {
		try {
			fileOutStream = new FileOutputStream(queueFile, false);
			queueOutStream = new ObjectOutputStream(fileOutStream);
			
			queueOutStream.writeObject(queue);
			queueOutStream.flush();
			
			queueOutStream.close();
			fileOutStream.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
